{                          MSIHELP.INC
                           MSINST 4.0
             Copyright (c) 1985, 87 by Borland International, Inc.            }

  procedure HelpInstall;
    {-Build an indexed help file for use by the editor}

  type
    {File position and block length of help sections}
    HelpIndexRec =
    record
      start : LongInt;
      Len : Integer;
    end;

    {File pointers to start of help for each command}
    HelpIndex = array[CommandType] of HelpIndexRec;
    TextFile = Text {[4096]} ;

  var
    helpf, f : file;
    textf : TextFile;
    index : HelpIndex;
    bytes : Word;

    procedure ScanFile(var textf : TextFile; var helpf : file);
      {-Scan text help, building index, assigning keystrokes, and formatting}
    label
      endofloop;
    const
      eolmark : string[1] = ^N;
      eopmark : string[1] = ^L;
      eosmark : string[1] = ^Z;
    var
      iline, oline : String255;
      ilen : Byte absolute iline;
      olen : Byte absolute oline;
      fpos : LongInt;        {Position in output file}
      inum : Integer;        {Line number of input}
      onum : Integer;        {Line number of current output page}
      cord : Integer;        {Ord of Current command}
      sectlen, maxlen : Integer; {Bytes per section}
      msg : VarString;
      spaces : Integer;

      procedure SectionOut(var onum : Integer; var fpos : LongInt);
        {-Insert section break in text stream}
      var
        bytes : Word;

      begin                  {SectionOut}
        {Insert section break}
        BlockWrite(helpf, eosmark[1], 1, bytes);
        if (bytes <> 1) or (IoResult <> 0) then
          HaltError('Error writing end of section to indexed help file');
        fpos := fpos+1;
        onum := 1;
      end;                   {SectionOut}

      procedure PageOut(var onum : Integer; var fpos : LongInt);
        {-Insert page break in text stream}
      var
        bytes : Word;

      begin                  {PageOut}
        {Insert page break}
        BlockWrite(helpf, eopmark[1], 1, bytes);
        if (bytes <> 1) or (IoResult <> 0) then
          HaltError('Error writing end of page to indexed help file');
        fpos := fpos+1;
        onum := 1;
      end;                   {PageOut}

      procedure LineOut(var oline : String255; var onum : Integer; var fpos : LongInt);
        {-Write out the current line, updating output line number and filepos}
      var
        olen : Byte absolute oline;
        wrappos : Byte;
        bytes : Word;

        function GetWrapPos(oline : String255) : Byte;
          {-Return position of last character to write in oline}
        var
          o : Byte;
          msg : VarString;

        begin                {GetWrapPos}
          o := Succ(HelpWidth);
          {Find rightmost blank inside of margin}
          while (o > 1) and (oline[o] <> ' ') do
            o := Pred(o);
          if o <= 1 then begin
            Str(inum, msg);
            HaltError('Line '+msg+' cannot be broken for wrap');
          end;
          {Find next non-blank to left}
          while (o > 1) and (oline[o] = ' ') do
            o := Pred(o);
          GetWrapPos := o;
        end;                 {GetWrapPos}

      begin                  {LineOut}

        {Get the length to write}
        if olen > HelpWidth then
          wrappos := GetWrapPos(oline)
        else
          wrappos := olen;

        {Write the text}
        BlockWrite(helpf, oline[1], wrappos, bytes);
        if (bytes <> wrappos) or (IoResult <> 0) then
          HaltError('Error writing text to indexed help file');

        {Write the end of line mark}
        BlockWrite(helpf, eolmark[1], 1, bytes);
        if (bytes <> 1) or (IoResult <> 0) then
          HaltError('Error writing end of line to indexed help file');

        {Update the file position}
        fpos := fpos+1+wrappos;

        {Update the line count}
        onum := Succ(onum);
        if onum > HelpHeight then
          PageOut(onum, fpos);

        {Delete text we just wrote}
        Delete(oline, 1, Succ(wrappos));

      end;                   {LineOut}

      procedure ExpandKeystrokes(var iline : String255);
        {-Scan looking for keystroke symbols, replacing them with keys}
      var
        ilen : Byte absolute iline;
        p : Byte;
        numstr, txt : VarString;
        Code, num : Integer;

      begin                  {ExpandKeystrokes}
        p := Pos(^p, iline);
        if p <> 0 then begin
          {Primary key sequence}
          p := 1;
          while p <= ilen do begin
            if iline[p] = ^p then begin
              Delete(iline, p, 1);
              {Get the key number}
              numstr := '';
              while (p <= ilen) and (iline[p] in ['0'..'9']) do begin
                numstr := numstr+iline[p];
                Delete(iline, p, 1);
              end;
              if numstr <> '' then begin
                {Convert to number}
                Val(numstr, num, Code);
                if Code <> 0 then
                  HaltError('illegal primary keystroke code '+numstr);
                {Get keystroke text representation}
                txt := TextRepresentation(Commands[CommandType(num)].main);
                if txt = '' then
                  txt := '(None)'
                else
                  txt := ^B+txt+^c;
                {Insert it in line}
                Insert(txt, iline, p);
              end;
            end;
            p := Succ(p);
          end;
        end;
        p := Pos(^S, iline);
        if p <> 0 then begin
          {Secondary key sequence}
          p := 1;
          while p <= ilen do begin
            if iline[p] = ^S then begin
              Delete(iline, p, 1);
              {Get the key number}
              numstr := '';
              while (p <= ilen) and (iline[p] in ['0'..'9']) do begin
                numstr := numstr+iline[p];
                Delete(iline, p, 1);
              end;
              if numstr <> '' then begin
                {Convert to number}
                Val(numstr, num, Code);
                if Code <> 0 then
                  HaltError('illegal secondary keystroke code '+numstr);
                {Get keystroke text representation}
                txt := TextRepresentation(Commands[CommandType(num)].alt);
                if txt = '' then
                  txt := '(None)'
                else
                  txt := ^B+txt+^c;
                {Insert it in line}
                Insert(txt, iline, p);
              end;
            end;
            p := Succ(p);
          end;
        end;
      end;                   {ExpandKeystrokes}

      function GetCommandCode(iline : String255) : Byte;
        {-Error check and return command code}
      var
        numstr : String255;
        ilen : Byte absolute iline;
        Code, cord : Integer;
        msg : VarString;

      begin                  {GetCommandCode}
        numstr := '';
        while (ilen > 0) and (iline[1] in ['0'..'9']) do begin
          numstr := numstr+iline[1];
          Delete(iline, 1, 1);
        end;
        if numstr = '' then begin
          Str(inum, msg);
          HaltError('missing command code on input line '+msg);
        end;
        Val(numstr, cord, Code);
        if Code <> 0 then begin
          Str(inum, msg);
          HaltError('illegal command code on input line '+msg);
        end;
        GetCommandCode := cord;
      end;                   {GetCommandCode}

      procedure WriteFunctionKeys(var onum : Integer; var fpos : LongInt);
        {-Write a map of the function keys in use}
      var
        s : Byte;
        cmdname : CommandName;
        oline : String255;
        olen : Byte absolute oline;
        k : keyrec;

        function KeyMatch(k : KeyString; var cmd : CommandName) : Boolean;
          {-Return True and a command name if keys matches an installed command}
        var
          c : CommandType;

        begin                {KeyMatch}
          c := CmdLeftChar;
          while c <= CmdNullMain do begin
            with Commands[c] do begin
              if k = main.keys then begin
                cmdname := name;
                KeyMatch := True;
                Exit;
              end;
              if k = alt.keys then begin
                cmdname := name;
                KeyMatch := True;
                Exit;
              end;
            end;
            c := Succ(c);
          end;
          KeyMatch := False;
        end;                 {KeyMatch}

      begin                  {WriteFunctionKeys}

        for s := 59 to 68 do
          {Normal function keys}
          if KeyMatch(#0+Chr(s), cmdname) then begin
            k.keys := #0+Chr(s);
            oline := ^B+TextRepresentation(k)+^c;
            while olen < 14 do
              oline := oline+' ';
            oline := oline+cmdname;
            while olen > 0 do
              LineOut(oline, onum, fpos);
          end;

        for s := 84 to 93 do
          {Shifted function keys}
          if KeyMatch(#0+Chr(s), cmdname) then begin
            k.keys := #0+Chr(s);
            oline := ^B+TextRepresentation(k)+^c;
            while olen < 14 do
              oline := oline+' ';
            oline := oline+cmdname;
            while olen > 0 do
              LineOut(oline, onum, fpos);
          end;

        for s := 94 to 103 do
          {Ctrl function keys}
          if KeyMatch(#0+Chr(s), cmdname) then begin
            k.keys := #0+Chr(s);
            oline := ^B+TextRepresentation(k)+^c;
            while olen < 14 do
              oline := oline+' ';
            oline := oline+cmdname;
            while olen > 0 do
              LineOut(oline, onum, fpos);
          end;

        for s := 104 to 113 do
          {Alt function keys}
          if KeyMatch(#0+Chr(s), cmdname) then begin
            k.keys := #0+Chr(s);
            oline := ^B+TextRepresentation(k)+^c;
            while olen < 14 do
              oline := oline+' ';
            oline := oline+cmdname;
            while olen > 0 do
              LineOut(oline, onum, fpos);
          end;

      end;                   {WriteFunctionKeys}

    begin                    {ScanFile}

      {First file position after the index}
      fpos := SizeOf(index);

      {Initialize}
      cord := 255;
      olen := 0;
      onum := 1;
      inum := 0;
      maxlen := 0;

      {Scan text file}
      while not(EoF(textf)) do begin

        {Get next line}
        ReadLn(textf, iline);
        if iline[1] = ^Z then {!!!!}
          goto endofloop;

        inum := Succ(inum);
        if (inum and 15) = 0 then begin
          Str(inum, msg);
          EdFastWrite(msg, 1, 30, LoColor);
        end;

        if ilen = 0 then begin

          {Write remainder of current line}
          while olen > 0 do
            LineOut(oline, onum, fpos);
          {Add a blank line}
          LineOut(oline, onum, fpos);

        end else

          {Check for column 1 commands}
          case iline[1] of

            '@' :
              {Start of new command}
              begin
                Delete(iline, 1, 1);
                while olen > 0 do
                  {Write out previous line}
                  LineOut(oline, onum, fpos);
                if cord <> 255 then begin
                  {End previous section}
                  SectionOut(onum, fpos);
                  {Store previous length}
                  sectlen := fpos-index[CommandType(cord)].start;
                  if sectlen > maxlen then
                    maxlen := sectlen;
                  index[CommandType(cord)].Len := sectlen;
                end;
                {Get new command number}
                cord := GetCommandCode(iline);
                {Store file index position}
                index[CommandType(cord)].start := fpos;
                {Ignore rest of line - can have comments there}
                ilen := 0;
              end;

            '&' :
              {Center this line, assumed to have no expanded keys}
              begin
                Delete(iline, 1, 1);
                spaces := (HelpWidth-ilen) shr 1;
                if spaces > 0 then begin
                  msg[0] := Chr(spaces);
                  FillChar(msg[1], spaces, ' ');
                  Insert(msg, iline, 1);
                end;
              end;

            ')' :
              {Force line break}
              begin
                Delete(iline, 1, 1);
                while olen > 0 do
                  LineOut(oline, onum, fpos);
              end;

            ']' :
              {Force page break}
              begin
                Delete(iline, 1, 1);
                while olen > 0 do
                  LineOut(oline, onum, fpos);
                if onum > 1 then
                  PageOut(onum, fpos);
              end;

            '|' :
              {Insert function key map}
              begin
                Delete(iline, 1, 1);
                while olen > 0 do
                  LineOut(oline, onum, fpos);
                WriteFunctionKeys(onum, fpos);
              end;

          end;

        if ilen > 0 then begin
          {Normal text line}
          {Expand keystrokes}
          ExpandKeystrokes(iline);
          {Attach to oline}
          if olen = 0 then
            oline := iline
          else
            oline := oline+' '+iline;

          {Wrap oline}
          while olen >= HelpWidth do
            LineOut(oline, onum, fpos);
        end;

      end;

endofloop:

      {End previous section}
      while olen > 0 do
        {Write out previous line}
        LineOut(oline, onum, fpos);
      SectionOut(onum, fpos);
      {Store previous length}
      sectlen := fpos-index[CommandType(cord)].start;
      if sectlen > maxlen then
        maxlen := sectlen;
      index[CommandType(cord)].Len := sectlen;

      RestoreScreen;
      {WriteLn('Longest help section: ', maxlen, ' bytes');}

    end;                     {ScanFile}

  begin                      {HelpInstall}

    {Assure input text help file found}
    if not(FileExists(TxtFile, f)) then
      HaltError('Text help file not found');
    Close(f);

    Assign(textf, TxtFile);
    Reset(textf);

    {Open indexed help file}
    Assign(helpf, HelpFile);
    Rewrite(helpf, 1);

    {Initialize help index}
    FillChar(index, SizeOf(index), 0);

    {Allocate file space for help index}
    BlockWrite(helpf, index, SizeOf(index), bytes);
    if (bytes <> SizeOf(index)) or (IoResult <> 0) then
      HaltError('Unable to write index to indexed help file');

    {Scan text file, building index, assigning keys, and formatting}
    ClrEol(1, 1, LoColor);
    EdFastWrite('Writing indexed Help file...', 1, 1, LoColor);
    ScanFile(textf, helpf);

    {Write help index}
    Seek(helpf, 0);
    BlockWrite(helpf, index, SizeOf(index), bytes);
    if (bytes <> SizeOf(index)) or (IoResult <> 0) then
      HaltError('Unable to write index to indexed help file');

    {Close up}
    Close(helpf);
    if IoResult <> 0 then
      HaltError('Error closing indexed help file');

  end;                       {HelpInstall}
